// Copyright 2018 RnD Center "ELVEES", JSC
#include <cassert>
#include <ctime>
#include <mutex>
#include <sstream>
#include <stdexcept>

#include "elcore-cl.hpp"

namespace {
    void ECL_CALLBACK MemoryDestructor(ecl_mem, void *user_data) { free(user_data); }
}

void _ecl_command_queue::PollThread() {
    do {
        boost::intrusive_ptr<ecl_command_queue_node> kernel_to_poll;
        {
            std::unique_lock<std::mutex> lock(m_commands_list_guard);
            while (m_submitted_kernels_list.empty() && ref_count() > 1) {
                m_commands_list_cond.wait(lock);
            }
            if (ref_count() == 1) break;
            assert(!m_submitted_kernels_list.empty());
            kernel_to_poll = m_submitted_kernels_list.front();
            m_submitted_kernels_list.pop_front();
        }
        int job_instance_fd = kernel_to_poll->m_kernel->m_job_instance.job_instance_fd;
        int status = ElcoreJobInstancePoll(job_instance_fd, m_device->m_common_fd);

        if (m_device->close_job(job_instance_fd) != 0)
            throw std::runtime_error("could not close job FD");
        {
            std::unique_lock<std::mutex> lock(m_commands_list_guard);
            if (kernel_to_poll == m_last_command_in_queue) m_last_command_in_queue = nullptr;
        }
        kernel_to_poll->m_event->set_command_status(status != ELCORE50_JOB_STATUS_SUCCESS
                                                        ? -1
                                                        : ECL_COMPLETE);
    } while (ref_count() > 1);
    release();
}

static ecl_int commandQueueCheckParams(ecl_context context, ecl_device_id device,
                                       const ecl_queue_properties* properties) {
    if (context == nullptr) return ECL_INVALID_CONTEXT;
    if (device == nullptr) return ECL_INVALID_DEVICE;

    ecl_bool found = ECL_FALSE;

    for (size_t i = 0; i < context->m_device_ids.size(); i++) {
        if (context->m_device_ids[i] == device) found = ECL_TRUE;
    }
    if (found == ECL_FALSE) return ECL_INVALID_DEVICE;

    return ECL_SUCCESS;
}

template<typename T>
static boost::intrusive_ptr<T>
createCommandQueue(ecl_context context, ecl_device_id device,
                   const ecl_queue_properties* properties, ecl_int &ret) {
    ecl_uint i = 0;
    ecl_command_queue_properties queue_props = 0;
    int queue_props_set = 0;
    if (properties) {
        while (properties[i]) {
            switch (properties[i]) {
                case ECL_QUEUE_PROPERTIES:
                    queue_props = (ecl_command_queue_properties)properties[i + 1];
                    queue_props_set = 1;
                    i += 2;
                    break;
                case ECL_QUEUE_SIZE:
                    ret = ECL_INVALID_QUEUE_PROPERTIES;
                    return nullptr;
                default:
                    ret = ECL_INVALID_VALUE;
                    return nullptr;
            }
        }
    }

    if (queue_props_set) {
        if (queue_props & (ECL_QUEUE_ON_DEVICE | ECL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE |
                           ECL_QUEUE_ON_DEVICE_DEFAULT)) {
            ret = ECL_INVALID_QUEUE_PROPERTIES;
            return nullptr;
        }

        const ecl_command_queue_properties valid_prop_flags =
            (ECL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE | ECL_QUEUE_PROFILING_ENABLE |
             ECL_QUEUE_ON_DEVICE | ECL_QUEUE_ON_DEVICE_DEFAULT);
        if (queue_props & (~valid_prop_flags)) {
            ret = ECL_INVALID_VALUE;
            return nullptr;
        }
    }

    boost::intrusive_ptr<T> command_queue(new T(device, context, queue_props), false);

    ret = ECL_SUCCESS;

    command_queue->add_ref();

    return command_queue;
}

extern "C" ECL_API_ENTRY ecl_command_queue ECL_API_CALL
eclCreateCommandQueueWithProperties(ecl_context context, ecl_device_id device,
                                    const ecl_queue_properties* properties,
                                    ecl_int* errcode_ret) ECL_API_SUFFIX__VERSION_2_0 {
    ecl_int ret;

    ret = commandQueueCheckParams(context, device, properties);
    if (ret != ECL_SUCCESS) {
        if (errcode_ret) *errcode_ret = ret;
        return nullptr;
    }

    boost::intrusive_ptr<_ecl_command_queue> command_queue;
    command_queue = createCommandQueue<_ecl_command_queue>(context, device, properties, ret);
    if (command_queue == nullptr || ret != ECL_SUCCESS) {
        if (errcode_ret) *errcode_ret = ret;
        return nullptr;
    }

    if (errcode_ret != nullptr) *errcode_ret = ECL_SUCCESS;

    return command_queue.get();
}

extern "C" ECL_API_ENTRY ecl_multi_command_queue ECL_API_CALL
eclCreateMultiCommandQueueWithProperties(ecl_context context, ecl_device_id device,
                                         const ecl_queue_properties* properties,
                                         ecl_uint num_spawn_devices,
                                         const ecl_device_id* spawn_device_list,
                                         ecl_int* errcode_ret) ECL_API_SUFFIX__VERSION_2_0 {
    ecl_int ret;

    ret = commandQueueCheckParams(context, device, properties);
    if (ret != ECL_SUCCESS) {
        if (errcode_ret) *errcode_ret = ret;
        return nullptr;
    }

    // check for duplicates in spawn_device_list[].
    for (std::size_t i = 0; i < context->m_device_ids.size(); i++) {
        int count = 0;
        for (size_t j = 0; j < num_spawn_devices; j++) {
            count += context->m_device_ids[i] == spawn_device_list[j];
        }
        // duplicate devices
        if (count > 1) {
            if (errcode_ret) *errcode_ret = ECL_INVALID_DEVICE;
            return nullptr;
        }
    }

    // check for invalid devices in spawn_device_list[].
    for (size_t i = 0; i < num_spawn_devices; i++) {
        bool found = 0;
        for (std::size_t j = 0; j < context->m_device_ids.size(); j++) {
            found |= context->m_device_ids[j] == spawn_device_list[i];
        }
        if (!found) {
            if (errcode_ret) *errcode_ret = ECL_INVALID_DEVICE;
            return nullptr;
        }
    }

    // check if `device` is not in spawn_device_list
    for (size_t i = 0; i < num_spawn_devices; i++) {
        if (device == spawn_device_list[i]) {
            if (errcode_ret) *errcode_ret = ECL_INVALID_DEVICE;
            return nullptr;
        }
    }

    size_t nc_size = (num_spawn_devices + 1) * SZ_1M;
    void *nc_data;
    kernel_spawn_internal_data *internal_data;

    if (posix_memalign((void **)&nc_data, 64, nc_size)) {
        if (errcode_ret != nullptr) {
            *errcode_ret = ECL_INVALID_BUFFER_SIZE;
        }
        return nullptr;
    }
    memset(nc_data, 0, nc_size);

    if (posix_memalign((void **)&internal_data, 64, sizeof(kernel_spawn_internal_data))) {
        if (errcode_ret != nullptr) {
            *errcode_ret = ECL_INVALID_BUFFER_SIZE;
        }
        free(nc_data);
        return nullptr;
    }

    boost::intrusive_ptr<_ecl_multi_command_queue> command_queue;
    command_queue = createCommandQueue<_ecl_multi_command_queue>(context, device, properties, ret);
    if (command_queue == nullptr || ret != ECL_SUCCESS) {
        if (errcode_ret) *errcode_ret = ret;
        return nullptr;
    }

    command_queue->m_nc_mem = eclCreateBuffer(context, ECL_MEM_USE_HOST_PTR, nc_size,
                                              nc_data, &ret);
    if (ret != ECL_SUCCESS) {
        if (errcode_ret != nullptr) {
            *errcode_ret = ret;
        }
        free(nc_data);
        free(internal_data);
        return nullptr;
    }
    eclSetMemObjectDestructorCallback(command_queue->m_nc_mem, MemoryDestructor, nc_data);

    command_queue->m_internal_mem =
        eclCreateBuffer(context, ECL_MEM_USE_HOST_PTR, sizeof(kernel_spawn_internal_data),
                        (void *)internal_data, &ret);
    if (ret != ECL_SUCCESS) {
        if (errcode_ret != nullptr) {
            *errcode_ret = ret;
        }
        free(internal_data);
        eclReleaseMemObject(command_queue->m_nc_mem);
        return nullptr;
    }
    eclSetMemObjectDestructorCallback(command_queue->m_internal_mem, MemoryDestructor,
                                      (void *)internal_data);


    ecl_command_queue queue;
    size_t i = 0;
    for (; i < num_spawn_devices; i++) {
        queue = eclCreateCommandQueueWithProperties(context, spawn_device_list[i],
                                                    properties, &ret);
        if (queue == nullptr || ret != ECL_SUCCESS) break;
        command_queue->m_spawn_queues.push_back(queue);
    }

    if (i != num_spawn_devices) {
        while (i > 0) {
            eclReleaseCommandQueue(command_queue->m_spawn_queues[--i]);
        }
        command_queue->m_spawn_queues.clear();
        eclReleaseMemObject(command_queue->m_internal_mem);
        eclReleaseMemObject(command_queue->m_nc_mem);
        if (errcode_ret != nullptr) *errcode_ret = ret;
        return nullptr;
    }

    if (errcode_ret != nullptr) *errcode_ret = ECL_SUCCESS;

    return command_queue.get();
}

extern "C" ECL_API_ENTRY ecl_int ECL_API_CALL
eclRetainCommandQueue(ecl_command_queue command_queue) ECL_API_SUFFIX__VERSION_1_0 {
    if (command_queue == nullptr) return ECL_INVALID_COMMAND_QUEUE;
    command_queue->add_ref();
    return ECL_SUCCESS;
}

extern "C" ECL_API_ENTRY ecl_int ECL_API_CALL
eclRetainMultiCommandQueue(ecl_multi_command_queue command_queue) ECL_API_SUFFIX__VERSION_1_0 {
    if (command_queue == nullptr) return ECL_INVALID_COMMAND_QUEUE;
    command_queue->add_ref();
    return ECL_SUCCESS;
}

extern "C" ECL_API_ENTRY ecl_int ECL_API_CALL
eclReleaseCommandQueue(ecl_command_queue command_queue) ECL_API_SUFFIX__VERSION_1_0 {
    if (command_queue == nullptr) return ECL_INVALID_COMMAND_QUEUE;
    command_queue->release();
    return ECL_SUCCESS;
}

extern "C" ECL_API_ENTRY ecl_int ECL_API_CALL
eclReleaseMultiCommandQueue(ecl_multi_command_queue command_queue) ECL_API_SUFFIX__VERSION_1_0 {
    if (command_queue == nullptr) return ECL_INVALID_COMMAND_QUEUE;
    command_queue->release();
    return ECL_SUCCESS;
}

extern "C" ECL_API_ENTRY ecl_int ECL_API_CALL
eclGetCommandQueueInfo(ecl_command_queue command_queue, ecl_command_queue_info param_name,
                       size_t param_value_size, void* param_value,
                       size_t* param_value_size_ret) ECL_API_SUFFIX__VERSION_1_0 {
    if (command_queue == nullptr) return ECL_INVALID_COMMAND_QUEUE;

    switch (param_name) {
        case ECL_QUEUE_CONTEXT:
            ECL_RETURN_GETINFO(ecl_context, command_queue->m_context.get());
        case ECL_QUEUE_DEVICE:
            ECL_RETURN_GETINFO(ecl_device_id, command_queue->m_device_id);
        case ECL_QUEUE_REFERENCE_COUNT:
            ECL_RETURN_GETINFO(ecl_uint, (ecl_uint)command_queue->ref_count());
        case ECL_QUEUE_PROPERTIES:
            ECL_RETURN_GETINFO(ecl_command_queue_properties, command_queue->m_properties);
    }
    return ECL_INVALID_VALUE;
}
